# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from __future__ import absolute_import

from contextlib import contextmanager
from os import rename, makedirs
from os.path import dirname, isdir
from tempfile import NamedTemporaryFile


@contextmanager
def atomic_writer(path, mode=0o644):
    """
    Write to path in an atomic operation. Auto creates any directories leading
    up to ``path``
    """
    d = dirname(path)
    if d:
        makedir(d)
    tmpfile = NamedTemporaryFile(delete=False, dir=d)
    yield tmpfile
    tmpfile.close()
    rename(tmpfile.name, path)


def makedir(path):
    """
    Create a directory at ``path``.
    Unlike ``os.makedirs`` don't raise an error if ``path`` already exists.
    """
    try:
        makedirs(path)
    except OSError:
        # Path already exists or cannot be created
        if not isdir(path):
            raise



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
#
"""
Common utilities for writing applications in fresco
"""
from __future__ import absolute_import

from fresco.exceptions import NotFound

__all__ = ['object_or_404']


def object_or_404(ob, exception=NotFound):
    """
    Return the value of ``ob`` if it is not None. Otherwise raise a NotFound
    exception.
    """
    if ob is None:
        raise exception()
    return ob


def fq_path(ob):
    """
    Return the fully qualified path of ``ob``, expected to be a function or
    method
    """
    name = getattr(ob, '__qualname__', None) or \
            getattr(ob, '__name__', None) or repr(ob)
    module = getattr(ob, '__module__', None) or ''
    return '{}.{}'.format(module, name)



# encoding: UTF-8
from fresco import FrescoApp, GET, Route, Response
from fresco.routeargs import GetArg, routearg
from fresco.subrequests import subrequest, subrequest_raw, subrequest_bytes

from mock import Mock
import pytest


class TestSubRequest(object):

    def test_it_raises_exception_if_decoding_impossible(self):

        def view_bytes():
            return Response(b'caf\xe9', content_type='text/plain')

        with FrescoApp().requestcontext():
            with pytest.raises(ValueError):
                subrequest(view_bytes)

    def test_it_decodes_response_content(self):

        def view_latin1():
            return Response(u'café'.encode('latin1'),
                            content_type='text/plain; charset=Latin-1')

        def view_utf8():
            return Response(u'café'.encode('UTF-8'),
                            content_type='text/plain; charset=UTF-8')

        def view_string():
            return Response(u'café', content_type='text/plain; charset=UTF-8')

        def view_list_of_strings():
            return Response([u'café'],
                            content_type='text/plain; charset=UTF-8')

        with FrescoApp().requestcontext():
            assert subrequest(view_utf8) == u'café'
            assert subrequest(view_latin1) == u'café'
            assert subrequest(view_string) == u'café'
            assert subrequest(view_list_of_strings) == u'café'

    def test_it_returns_a_markup_string(self):

        def view_html():
            return Response(u'<html>',
                            content_type='text/html; charset=UTF-8')

        def view_text():
            return Response(u'text',
                            content_type='text/plain; charset=UTF-8')

        with FrescoApp().requestcontext():
            assert hasattr(subrequest(view_html), '__html__')
            assert not hasattr(subrequest(view_text), '__html__')

    def test_it_returns_raw_response(self):
        r = Response('foo')
        with FrescoApp().requestcontext():
            assert subrequest_raw(lambda: r) is r

    def test_it_returns_content_as_bytes(self):
        view = lambda: Response(u'café'.encode('latin1'),
                                content_type='text/plain; charset=Latin-1')
        with FrescoApp().requestcontext():
            assert subrequest_bytes(view) == b'caf\xe9'

    def test_it_calls_response_onclose(self):
        m = Mock()
        r = Response(onclose=m)
        with FrescoApp().requestcontext():
            assert subrequest(lambda: r) == u''
        assert m.call_count == 1

    def test_viewspec_resolves(self):
        def view():
            return Response('foo')

        app = FrescoApp()
        app.route('/view', GET, view, name='named view')
        with app.requestcontext():
            assert subrequest('named view') == 'foo'

    def test_viewspec_uses_routeargs(self):

        def view(a, b):
            return Response(['*' + a + b + '*'])

        app = FrescoApp()
        app.route('/view', GET, view,
                  a=GetArg(),
                  b=routearg(lambda r: r.method))

        with app.requestcontext('/view?a=foo'):
            assert subrequest(view, _resolve=True) == '*fooGET*'
            assert subrequest(view, _resolve=True, a='bar', b='baz') \
                    == '*barbaz*'

    def test_viewspec_resolves_dynamic(self):

        class Views(object):

            def __init__(self, a):
                self.a = a

            def view(self, b):
                return Response('*' + self.a + b + '*')

            __routes__ = [Route('/<b:str>', GET, 'view', name='named view')]

        app = FrescoApp()
        app.delegate('/<a:str>', Views, dynamic=True, name='named collection')

        with app.requestcontext():
            assert subrequest(
                'named collection:named view', a='foo', b='bar') == '*foobar*'

    def test_it_does_a_full_request(self):

        request_hook = Mock(return_value=None)
        view = Mock(return_value=Response())

        app = FrescoApp()
        app.route('/view', GET, view)
        app.process_request(request_hook)
        with app.requestcontext():
            subrequest(view)
            assert view.call_count == 1
            assert request_hook.call_count == 0

            subrequest(view, _full=True)
            assert view.call_count == 2
            assert request_hook.call_count == 1



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
# flake8: noqa
from __future__ import absolute_import

from sys import version_info

PY3 = version_info >= (3, 3)

if PY3:
    string_types = (str,)
    ustr = str
else:
    string_types = (basestring,)
    ustr = unicode

if PY3:
    from threading import get_ident

else:
    from thread import get_ident

if PY3:
    from urllib.parse import (quote, quote_plus, unquote, urlparse, urlunparse,
                              ParseResult)

else:
    from urllib import quote, quote_plus, unquote
    from urlparse import urlparse, urlunparse, ParseResult


if PY3:
    def reraise(exc_info):
        raise exc_info[1].with_traceback(exc_info[2])

else:
    exec('def reraise(exc_info): raise exc_info[0], exc_info[1], exc_info[2]')

__all__ = 'PY3', 'string_types', 'ustr',\
          'get_ident', 'quote', 'quote_plus', 'unquote', 'urlparse', \
          'ParseResult', 'reraise'



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
"""
Utilities functions for URL building and manipulation
"""

from __future__ import absolute_import

import posixpath
from fresco.compat import quote_plus, urlparse, urlunparse, ustr, string_types

import fresco
from fresco.multidict import MultiDict

__all__ = 'join_path', 'url_join', 'strip_trailing_slashes', 'normpath'


def join_path(a, b):
    """
    Join two URL path segments together.
    Unlike ``os.path.join``, this removes leading slashes before joining the
    paths, so absolute paths are appended as if they were relative.

    Synopsis::

        >>> join_path('/', '/abc')
        '/abc'

        >>> join_path('/', '/abc/')
        '/abc/'

        >>> join_path('/foo', '/bar/')
        '/foo/bar/'

        >>> join_path('/foo/', '/bar/')
        '/foo/bar/'

        >>> join_path('/foo', 'bar')
        '/foo/bar'
    """

    if not b:
        return a

    while a and a[-1] == '/':
        a = a[:-1]

    while b and b[0] == '/':
        b = b[1:]

    return a + '/' + b


def url_join(base, rel):
    """
    Examples::

        >>> from fresco.util.urls import url_join
        >>> url_join('http://example.org/', 'http://example.com/')
        'http://example.com/'
        >>> url_join('http://example.com/', '../styles/main.css')
        'http://example.com/styles/main.css'
        >>> url_join('http://example.com/subdir/', '../styles/main.css')
        'http://example.com/styles/main.css'
        >>> url_join('http://example.com/login', '?error=failed+auth')
        'http://example.com/login?error=failed+auth'
        >>> url_join('http://example.com/login', 'register')
        'http://example.com/register'
    """
    prel = urlparse(rel)

    # Link is already absolute, return it unchanged
    if prel.scheme:
        return rel

    pbase = urlparse(base)
    path = pbase.path
    if prel.path:
        path = normpath(posixpath.join(posixpath.dirname(pbase.path),
                                       prel.path))

    return urlunparse((
        pbase.scheme,
        pbase.netloc,
        path,
        prel.params,
        prel.query,
        prel.fragment
    ))


def strip_trailing_slashes(path):
    """\
    Remove trailing slashes from the given path.

    :param path: a path, eg ``'/foo/bar/'``
    :return: The path with trailing slashes removed, eg ``'/foo/bar'``
    """
    while path and path[-1] == '/':
        path = path[:-1]
    return path


def normpath(path):
    """
    Return ``path`` normalized to remove '../' etc.

    This differs from ``posixpath.normpath`` in that:

    * trailing slashes are preserved
    * multiple consecutive slashes are always condensed to a single slash

    Examples::

        >>> normpath('/hello/../goodbye')
        '/goodbye'
        >>> normpath('//etc/passwd')
        '/etc/passwd'
        >>> normpath('../../../../etc/passwd')
        'etc/passwd'
        >>> normpath('/etc/passwd/')
        '/etc/passwd/'
    """
    segments = path.split('/')
    newpath = []
    last = len(segments) - 1
    for pos, seg in enumerate(segments):
        if seg == '.':
            seg = ''

        if seg == '':
            allow_empty = (
                pos == 0
                or pos == last and newpath and newpath[-1] != ''
                or pos == last and newpath == ['']
            )
            if not allow_empty:
                continue

        if seg == '..':
            if newpath and newpath != ['']:
                newpath.pop()
            continue

        newpath.append(seg)

    return '/'.join(newpath)


def make_query(data=None, separator=';', charset=None, **kwargs):
    """\
    Return a query string formed from the given dictionary data.

    Note that the pairs are separated using a semicolon, in accordance with
    `the W3C recommendation
<http://www.w3.org/TR/1999/REC-html401-19991224/appendix/notes.html#h-B.2.2>`_

    If no encoding is given, unicode values are encoded using the character set
    specified by ``fresco.DEFAULT_CHARSET``.

    Basic usage::

        >>> make_query({'search': 'foo bar', 'type': u'full'})
        'search=foo+bar;type=full'

        >>> make_query(search='foo', type=u'quick')
        'search=foo;type=quick'

    Changing the separator to '&'::

        >>> make_query({'search': 'foo', 'type': u'full'}, separator='&')
        'search=foo&type=full'

    Multiple values can be provided per key::

        >>> make_query(search=['foo', u'bar'])
        'search=foo;search=bar'

    Exclude values by setting them to ``None``::

        >>> make_query(x=1, y=None)
        'x=1'

    :param data: data for query string
    :param separator: separator used between each key value pair
    :param charset: encoding to be used for unicode values
    :rtype: str
    """
    if isinstance(data, MultiDict):
        items = list(data.allitems())
    elif isinstance(data, dict):
        items = list(data.items())
    elif data is None:
        items = []
    else:
        items = list(data)
    items += list(kwargs.items())

    if charset is None:
        charset = fresco.DEFAULT_CHARSET

    items = _repeat_keys(items)
    items = ((k, v) for k, v in items if v is not None)
    return separator.join(_qs_frag(k, v, charset=charset) for k, v in items)


def _qs_frag(key, value, charset=None):
    """\
    Return a fragment of a query string in the format 'key=value'::

        >>> _qs_frag('search-by', 'author, editor')
        'search-by=author%2C+editor'

    If no encoding is specified, unicode values are encoded using the character
    set specified by ``fresco.DEFAULT_CHARSET``.

    :rtype: str
    """
    if charset is None:
        charset = fresco.DEFAULT_CHARSET

    key = ustr(key).encode(charset)
    value = ustr(value).encode(charset)

    return quote_plus(key) + '=' + quote_plus(value)


def _repeat_keys(iterable):
    """
    Return a list of ``(key, scalar_value)`` tuples given an iterable
    containing ``(key, iterable_or_scalar_value)``.

    Example::

        >>> list(
        ...     _repeat_keys([('a', 'b')])
        ... )
        [('a', 'b')]
        >>> list(
        ...     _repeat_keys([('a', ['b', 'c'])])
        ... )
        [('a', 'b'), ('a', 'c')]
    """

    for key, value in iterable:
        if isinstance(value, string_types):
            value = [value]
        else:
            try:
                value = iter(value)
            except TypeError:
                value = [value]

        for subvalue in value:
            yield key, subvalue



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from __future__ import absolute_import

from fresco.response import Response

__all__ = 'ResponseException', 'NotFound', 'Unauthorized',\
          'Forbidden', 'BadRequest', 'Redirect', 'RedirectTemporary',\
          'RedirectPermanent'


class ResponseException(Exception):
    """\
    An exception class with an associated :class:`fresco.response.Response`
    instance that will render a suitable error page.
    """
    response = None

    #: If true, the error response will be returned immediately without trying
    #: any other routes.
    is_final = True

    def __init__(self, content=None, **kwargs):
        self.is_final = kwargs.pop('final', True)
        if content:
            kwargs['content'] = content
        if kwargs:
            self.response = self.response.replace(**kwargs)


class NotFound(ResponseException):
    """\
    Raised when a view needs to signal a 404 not found condition.
    """
    response = Response.not_found()

    def __init__(self, content=None, **kwargs):
        """
        ``final`` is false by default, meaning other routes will be offered the
        opportunity to handle the request. Only after other routes have been
        tried will an error response be returned.
        """
        kwargs.setdefault('final', False)
        super(NotFound, self).__init__(content, **kwargs)


class BadRequest(ResponseException):
    """\
    Return a 400 Bad Request response
    """
    response = Response(['Bad request'], status=400)


class Unauthorized(ResponseException):
    """\
    Return a 401 Unauthorized response.

    Use this when you want to flag that the user does not have permission to
    access a resource but may be able to gain access, eg by logging in.
    """
    response = Response(['Unauthorized'], status=401)


class Forbidden(ResponseException):
    """\
    Return a 403 Forbidden response

    Use this when you want to flag that the user does not have permission to
    access a resource and that there is no possiblity to do so (eg they are
    already logged in, but their account does not have the right access
    permissions)
    """
    response = Response(['Forbidden'], status=403)


class RedirectTemporary(ResponseException):
    """\
    Return a 302 Found response. Example:

        raise RedirectTemporary('http://example.org/')
    """
    def __init__(self, *args, **kwargs):
        self.response = Response.redirect(*args, **kwargs)

Redirect = RedirectTemporary


class RedirectPermanent(ResponseException):
    """\
    Return a 301 Moved Permanently response. Example:

        raise RedirectPermanent('http://example.org/')

    """
    def __init__(self, *args, **kwargs):
        self.response = Response.redirect_permanent(*args, **kwargs)


class RequestParseError(BadRequest):
    """\
    An error was encountered while parsing the HTTP request.
    """


class MethodNotAllowed(ResponseException):
    """\
    The URL is valid, but does not respond to the requested method
    """
    def __init__(self, valid_methods, **kwargs):
        self.response = Response.method_not_allowed(valid_methods)
        super(MethodNotAllowed, self).__init__(**kwargs)


class TooManyRequests(ResponseException):
    """
    Indicates that the user has sent too many
    requests in a given amount of time ("rate limiting").

    The response representations SHOULD include details explaining the
    condition, and MAY include a Retry-After header indicating how long
    to wait before making a new request.
    """
    response = Response(['Too many requests'], status=429)


class MissingRouteArg(BadRequest):
    """\
    A request argument was expected but not supplied.
    """



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from __future__ import absolute_import


__version__ = '0.7.0'

DEFAULT_CHARSET = 'UTF-8'

from .request import *  # noqa
from .requestcontext import context  # noqa
from .response import *  # noqa
from .core import *  # noqa
from .routing import *  # noqa
from .routeargs import *  # noqa
from .subrequests import *  # noqa

from .util.common import *  # noqa



from functools import wraps


class classorinstancemethod(object):

    def __init__(self, func):
        self.func = func

    def __get__(self, ob=None, obtype=None):

        @wraps(self.func)
        def decorated(*args, **kwargs):
            selfarg = ob if ob is not None else obtype
            return self.func(selfarg, *args, **kwargs)

        return decorated



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from __future__ import absolute_import

import pytest

from fresco.util import security


class TestSecurity(object):

    def test_check_equal_constant_time_returns_equality(self):
        assert security.check_equal_constant_time('', '') is True
        assert security.check_equal_constant_time('abcabcabc', 'abcabcabc') \
                is True

    def test_check_equal_constant_time_returns_inequality(self):
        assert security.check_equal_constant_time(' ', '') is False
        assert security.check_equal_constant_time('abcabcabc', 'abcabcabx') \
                is False
        assert security.check_equal_constant_time('abcabcabc', 'abcabcabx') \
                is False
        assert security.check_equal_constant_time('abcabcabc', 'abcabcabde') \
                is False
        assert security.check_equal_constant_time('abcabcabc', 'abcabcab') \
                is False

    def test_check_equal_constant_time_raises_an_error_on_non_strings(self):
        with pytest.raises(TypeError):
            security.check_equal_constant_time(None, None)



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from datetime import tzinfo, datetime, timedelta
from fresco.cookie import Cookie


class FakeTZInfo(tzinfo):

    def utcoffset(self, dt):
        return timedelta(seconds=3600)

    def dst(self, dt):
        return timedelta(seconds=0)

    def tzname(self, dt):
        return 'ZZZ'


class TestCookie(object):

    def test_basic_cookie(self):
        """Simple cookie with just a field and value"""
        c = Cookie('key', 'value')
        assert str(c) == 'key=value;Path=/'

    def test_nopath_cookie(self):
        """Simple cookie with just a field and value"""
        c = Cookie('key', 'value', path=None)
        assert str(c) == 'key=value'

    def test_httponly_cookie(self):
        """Cookie as HttpOnly"""
        c = Cookie('key', 'value', httponly=True)
        assert str(c) == 'key=value;Path=/;HttpOnly'

    def test_secure_cookie(self):
        """Cookie as secure cookie"""
        c = Cookie('key', 'value', secure=True)
        assert str(c) == 'key=value;Path=/;Secure'

    def test_secure_and_httponly(self):
        """Cookie as both secure and httponly"""
        c = Cookie('key', 'value', secure=True, httponly=True)
        assert str(c) == 'key=value;Path=/;Secure;HttpOnly'

    def test_dates_are_rfc_formatted(self):
        c = Cookie('a', 'b', expires=datetime(2001, 1, 1, 12))
        assert 'Expires=Mon, 01 Jan 2001 12:00:00 GMT;' in str(c)
        c = Cookie('a', 'b', expires=datetime(2001, 1, 1, 12,
                                              tzinfo=FakeTZInfo()))
        assert 'Expires=Mon, 01 Jan 2001 11:00:00 GMT;' in str(c)



# encoding=UTF-8
# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from __future__ import absolute_import

from mock import Mock
import pytest

from fresco import FrescoApp, GET
from fresco.cookie import Cookie
from fresco.response import Response
from fresco.routing import RouteNotFound


class TestCannedResponses(object):

    def test_redirect(self):
        with FrescoApp().requestcontext():
            r = Response.redirect('/new_location')
        assert b'http://localhost/new_location' in b''.join(r.content_iterator)
        assert r.get_header('Location') == 'http://localhost/new_location'
        assert r.status == '302 Found'

    def test_redirect_temporary(self):
        with FrescoApp().requestcontext():
            r = Response.redirect_temporary('/new_location')
        assert b'http://localhost/new_location' in b''.join(r.content_iterator)
        assert r.get_header('Location') == 'http://localhost/new_location'
        assert r.status == '302 Found'

    def test_redirect_permanent(self):
        with FrescoApp().requestcontext():
            r = Response.redirect_permanent('/new_location')
        assert b'http://localhost/new_location' in b''.join(r.content_iterator)
        assert r.get_header('Location') == 'http://localhost/new_location'
        assert r.status == '301 Moved Permanently'

    def test_redirect_to_view(self):
        app = FrescoApp()
        view = Mock(return_value=Response())
        app.route('/arfle/<barfle:int>', GET, view)
        with app.requestcontext():
            r = Response.redirect(view, barfle=42)
        assert r.get_header('Location') == 'http://localhost/arfle/42'

    def test_redirect_to_view_spec(self):
        app = FrescoApp()
        view = Mock(return_value=Response())
        app.route('/arfle/<barfle:int>', GET, view, name='gloop')
        with app.requestcontext():
            r = Response.redirect('gloop', barfle=42)
        assert r.get_header('Location') == 'http://localhost/arfle/42'

    def test_redirect_to_simple_string(self):
        with FrescoApp().requestcontext():
            r = Response.redirect('gloop')
        assert r.get_header('Location') == 'http://localhost/gloop'

    def test_redirect_raises_RouteNotFound(self):
        view = Mock()
        with FrescoApp().requestcontext():
            with pytest.raises(RouteNotFound):
                Response.redirect('gloop', arfle='a')
            with pytest.raises(RouteNotFound):
                Response.redirect(view)

    def test_not_found(self):
        with FrescoApp().requestcontext():
            r = Response.not_found()
        assert b'not found' in b''.join(r.content_iterator).lower()
        assert r.status_code == 404

    def test_method_not_allowed(self):
        with FrescoApp().requestcontext():
            r = Response.method_not_allowed(valid_methods=('PUT', 'DELETE'))
        assert b'not allowed' in b''.join(r.content_iterator).lower()
        assert r.status == '405 Method Not Allowed'
        assert r.get_header('Allow') == 'PUT,DELETE'

    def test_meta_refresh(self):
        with FrescoApp().requestcontext():
            r = Response.meta_refresh('/next_page')
        content = b''.join(r.content_iterator)
        assert b'<a href="http://localhost/next_page">' in content
        assert b'<meta http-equiv="refresh" content="0;'\
               b' url=http://localhost/next_page">' in content
        assert r.status == '200 OK'
        assert r.get_header('Content-Type') == 'text/html'

    def test_json(self):
        with FrescoApp().requestcontext():
            r = Response.json({'arfle': 'barfle'},
                              status=201,
                              indent=4,
                              separators=(',', ':\t'))
        assert r.status == '201 Created'
        assert r.get_header('Content-Type') == 'application/json'
        assert b''.join(r.content_iterator) == b'{\n    "arfle":\t"barfle"\n}'


class TestResponse(object):

    def test_unknown_response_header_from_python_symbol(self):
        r = Response(x_myheader='Foo', content_type=None)
        assert r.headers == [('X-Myheader', 'Foo')]

    def test_standard_headers_capitalized_from_symbol(self):
        r = Response(etag='foo')
        assert r.headers == [('ETag', 'foo')]

        r = Response(x_ua_compatible='foo')
        assert r.headers == [('X-UA-Compatible', 'foo')]

    def test_content_type_added_only_if_content(self):
        r = Response(content=[])
        assert 'Content-Type' in dict(r.headers)

        r = Response()
        assert 'Content-Type' not in dict(r.headers)

        r = Response(content=None)
        assert 'Content-Type' not in dict(r.headers)

    def test_add_response_header(self):
        r = Response(x_myheader='Foo')
        r = r.add_header('X-Myheader', 'Bar')
        assert r.headers == [('X-Myheader', 'Foo'), ('X-Myheader', 'Bar')]

    def test_set_content_type_header(self):
        r = Response(content_type='text/rtf; charset=ISO-8859-1')
        assert r.get_header('Content-Type') == 'text/rtf; charset=ISO-8859-1'

    def test_set_cookie(self):
        r = Response(set_cookie=Cookie(name='fruit', value='banana'))
        assert r.get_header('Set-Cookie') == 'fruit=banana;Path=/'

    def test_delete_cookie(self):
        r = Response().delete_cookie(name='fruit')
        assert r.get_header('Set-Cookie') == (
            'fruit=;Expires=Thu, 01 Jan 1970 00:00:00 GMT;Max-Age=0;Path=/')

    def test_response_get_headers(self):

        r = Response(['whoa nelly!'],
                     content_type='text/plain',
                     x_test_header=["1", "2"])
        assert r.get_header('content-type') == 'text/plain'
        assert r.get_header('x-test-header') == '1,2'
        assert r.get_header('X-Test-Header') == '1,2'
        assert r.get_header('x-does-not-exist', 'boo!') == 'boo!'

    def test_response_buffered_sets_content_length(self):
        r = Response(['whoa nelly!']).buffered()
        assert r.get_header('content-length') == '11'

        r = Response([u'whoa nélly!']).buffered()
        assert r.get_header('content-length') == '12'

    def test_it_encodes_a_unicode_string_according_to_content_type(self):
        r = Response(u'café', content_type='text/plain; charset=utf-8')
        assert list(r({}, Mock())) == [u'café'.encode('utf-8')]
        r = Response(u'café', content_type='text/plain; charset=latin-1')
        assert list(r({}, Mock())) == [u'café'.encode('latin-1')]


class TestAddVary(object):

    def test_it_adds_a_vary_header(self):
        r = Response(content_type='text/plain').add_vary('Accept-Encoding')
        assert ('Vary', 'Accept-Encoding') in r.headers

    def test_it_extends_an_existing_header(self):
        r = Response(content_type='text/plain')\
                .add_vary('Accept-Encoding')\
                .add_vary('Cookie')
        assert 'Cookie' in r.get_header('vary')
        assert 'Accept-Encoding' in r.get_header('vary')



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from __future__ import absolute_import

from unicodedata import normalize, combining


def strip_accents(s):
    """
    Return ``s`` with all accents stripped
    """
    return ''.join(c
                   for c in normalize('NFD', s)
                   if not combining(c))


def plural(quantity, one, multiple, zero=None, two=None):
    """
    Return a singular or plural strings for ``quantity``.

    Synopsis::

        >>> plural(1, '%d item', '%d items')
        '1 item'
        >>> plural(2, '%d item', '%d items')
        '2 items'

    Exceptions for zero and 2 quantities can be specified::

        >>> from functools import partial
        >>> p = partial(plural,
        ...     zero='what bottles?',
        ...     one='one bottle left',
        ...     two='a pair of bottles, hanging on the wall',
        ...     multiple='%d green bottles, hanging on the wall',
        ... )
        >>> for bottles in range(10, -1, -1):
        ...     print p(bottles)
        ...
        10 green bottles, hanging on the wall
        9 green bottles, hanging on the wall
        8 green bottles, hanging on the wall
        7 green bottles, hanging on the wall
        6 green bottles, hanging on the wall
        5 green bottles, hanging on the wall
        4 green bottles, hanging on the wall
        3 green bottles, hanging on the wall
        a pair of bottles, hanging on the wall
        one bottle left
        what bottles?
    """
    if zero is None:
        zero = multiple
    if two is None:
        two = multiple
    if quantity == 0:
        text = zero
    elif quantity == 1:
        text = one
    elif quantity == 2:
        text = two
    else:
        text = multiple

    try:
        return text % (quantity,)
    except TypeError:
        return text



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
"""
Utilities for dealing with static file content
"""
from __future__ import absolute_import

import mimetypes
import os
from email.utils import formatdate, parsedate_tz, mktime_tz

from fresco import context
from fresco.response import Response, STATUS_OK, STATUS_NOT_MODIFIED
from fresco.util.wsgi import ClosingIterator

__all__ = 'serve_static_file',


def serve_static_file(path, content_type=None, bufsize=8192, **kwargs):
    """
    Serve a static file located at ``path``.

    :returns: A :class:`~fresco.response.Response` object for the file at
              ``path``
    """

    try:
        mtime = os.path.getmtime(path)
    except OSError:
        return Response.not_found()

    request = context.request
    mod_since = request.get_header('if-modified-since')
    if mod_since is not None:
        try:
            mod_since = mktime_tz(parsedate_tz(mod_since))
        except (TypeError, OverflowError):
            return Response.bad_request()
        if int(mtime) <= int(mod_since):
            return Response(status=STATUS_NOT_MODIFIED)

    if content_type is None:
        content_type = mimetypes.guess_type(path)[0]
        if content_type is None:
            content_type = 'application/octet-stream'

    file_wrapper = request.environ.get('wsgi.file_wrapper')
    if file_wrapper is not None:
        content_iterator = lambda f: file_wrapper(f, bufsize)
    else:
        content_iterator = lambda f: \
                ClosingIterator(iter(lambda: f.read(bufsize), ''), f.close)

    try:
        _file = open(path, 'rb')
    except IOError:
        return Response.forbidden()

    return Response(
        status=STATUS_OK,
        content_length=str(os.path.getsize(path)),
        last_modified=formatdate(mtime, localtime=False, usegmt=True),
        content_type=content_type,
        content=content_iterator(_file),
        passthrough=True,
        **kwargs)



# encoding=UTF-8
# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from __future__ import absolute_import

from fresco import FrescoApp, Response
from fresco.util.wsgi import (ClosingIterator,
                              StartResponseWrapper,
                              getenv,
                              setenv)


class Counter(object):

    value = 0

    def inc(self):
        self.value += 1


def start_response(status, headers, exc_info=None):
    pass


class _TestException(Exception):
    pass


class TestClosingIterator(object):

    def app(self, environ, start_response):
        start_response('200 OK', [('Content-Type: text/plain')])
        yield "Foo"
        yield "Bar"

    def test_close_called_after_iterator_finished(self):
        count = Counter()
        environ = {}

        result = self.app(environ, start_response)
        result = ClosingIterator(result, count.inc)
        assert count.value == 0
        try:
            list(result)
        finally:
            result.close()
        assert count.value == 1

    def test_multiple_close_functions_called(self):
        count1 = Counter()
        count2 = Counter()
        environ = {}

        result = self.app(environ, start_response)
        result = ClosingIterator(result, count1.inc, count2.inc)
        assert count1.value == 0
        assert count2.value == 0
        try:
            list(result)
        finally:
            result.close()
        assert count1.value == 1
        assert count2.value == 1


class TestStartResponseWrapper(object):

    def test_write(self):

        def wsgiapp(environ, start_response):
            start_response = StartResponseWrapper(start_response)
            write = start_response('200 OK', [('Content-Type', 'text/plain')])

            write(b'cat')
            write(b'sat')
            write2 = start_response.call_start_response()
            write2(b'mat')
            return []

        with FrescoApp().requestcontext() as c:
            r = Response.from_wsgi(wsgiapp,
                                   c.request.environ,
                                   lambda status, headers: None)
            assert b''.join(r.content) == b'catsatmat'


class TestEnvironGetSet(object):

    wsgistr = u'ø'.encode('UTF-8')
    if str is not bytes:
        wsgistr = wsgistr.decode('ISO-8859-1')

    def test_getenv_returns_bytes(self):
        assert getenv({'x': self.wsgistr}, 'x') == u'ø'.encode('UTF-8')

    def test_setenv_sets_str(self):
        env = {}
        setenv(env, 'x', u'ø'.encode('UTF-8'))
        assert env['x'] == self.wsgistr



from collections import OrderedDict, MutableMapping


class LRUCache(MutableMapping):
    """
    An LRUCache implementation backed by collections.OrderedDict
    """

    def __init__(self, max_size):
        self._cache = OrderedDict()
        self.max_size = max_size

    @property
    def max_size(self):
        return self._max_size

    @max_size.setter
    def max_size(self, n):
        for i in range(len(self._cache) - n):
            self._cache.popitem(False)
        self._max_size = n

    def __setitem__(self, key, value):
        if len(self._cache) >= self._max_size:
            self._cache.popitem(last=False)
        self._cache[key] = value

    if hasattr(OrderedDict, 'move_to_end'):
        def __getitem__(self, key):
            value = self._cache[key]
            self._cache.move_to_end(key)
            return value

    else:
        def __getitem__(self, key):
            value = self._cache[key] = self._cache.pop(key)
            return value

    def __delitem__(self, key):
        self._cache.__delitem__[key]

    def __iter__(self):
        return self._cache.__iter__()

    def __reversed__(self):
        return self._cache.__reversed__()

    def __len__(self):
        return self._cache.__len__()

    def __contains__(self, key):
        return self._cache.contains(key)

    def __getattr__(self, attr):
        return getattr(self._cache, attr)

    def update(self, other):
        MutableMapping.update(self, other)

    def copy(self):
        copy = self.__class__()
        copy._cache = self._cache.copy()
        return copy



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
"""
Utilities for working with data on the HTTP level
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from binascii import hexlify
from collections import namedtuple
from email.message import Message
from email.header import Header
from io import BytesIO
from itertools import chain
import os
import re
try:
    from email.parser import BytesParser
except ImportError:
    # python 2
    from email.parser import Parser as BytesParser  # NOQA

try:
    from urllib.parse import unquote_plus
except ImportError:
    from urllib import unquote_plus  # NOQA
from shutil import copyfileobj

import fresco
from fresco.compat import PY3, string_types, ustr
from fresco.exceptions import RequestParseError
from fresco.util.io import (ExpandableOutput, SizeLimitedInput, PutbackInput,
                            DelimitedInput)
from fresco.util.wsgi import unicode_to_environ

KB = 1024
MB = 1024 * KB

#: Data chunk size to read from the input stream (wsgi.input)
CHUNK_SIZE = 8 * KB

#: Allowed character encodings.
#: This list is based on the list of standard encodings here:
#:
#:     https://docs.python.org/3.4/library/codecs.html
#:
#: Python specific encodings and binary transformations (eg zlib decompression)
#: are excluded to avoid zip bomb style attacks.
ALLOWED_ENCODINGS = set([
    'ascii', '646', 'us-ascii',
    'big5', 'big5-tw', 'csbig5',
    'big5hkscs', 'big5-hkscs', 'hkscs',
    'cp037', 'IBM037', 'IBM039',
    'cp273', '273', 'IBM273', 'csIBM273',
    'cp424', 'EBCDIC-CP-HE', 'IBM424',
    'cp437', '437', 'IBM437',
    'cp500', 'EBCDIC-CP-BE', 'EBCDIC-CP-CH', 'IBM500',
    'cp720', 'Arabic',
    'cp737', 'Greek',
    'cp775', 'IBM775',
    'cp850', '850', 'IBM850',
    'cp852', '852', 'IBM852',
    'cp855', '855', 'IBM855',
    'cp856', 'Hebrew',
    'cp857', '857', 'IBM857',
    'cp858', '858', 'IBM858',
    'cp860', '860', 'IBM860',
    'cp861', '861', 'CP-IS', 'IBM861',
    'cp862', '862', 'IBM862',
    'cp863', '863', 'IBM863',
    'cp864', 'IBM864',
    'cp865', '865', 'IBM865',
    'cp866', '866', 'IBM866',
    'cp869', '869', 'CP-GR', 'IBM869',
    'cp874', 'Thai',
    'cp875', 'Greek',
    'cp932', '932', 'ms932', 'mskanji', 'ms-kanji',
    'cp949', '949', 'ms949', 'uhc',
    'cp950', '950', 'ms950',
    'cp1006', 'Urdu',
    'cp1026', 'ibm1026',
    'cp1125', '1125', 'ibm1125', 'cp866u', 'ruscii',
    'cp1140', 'ibm1140',
    'cp1250', 'windows-1250',
    'cp1251', 'windows-1251',
    'cp1252', 'windows-1252',
    'cp1253', 'windows-1253',
    'cp1254', 'windows-1254',
    'cp1255', 'windows-1255',
    'cp1256', 'windows-1256',
    'cp1257', 'windows-1257',
    'cp1258', 'windows-1258',
    'cp65001',
    'euc_jp', 'eucjp', 'ujis', 'u-jis',
    'euc_jis_2004', 'jisx0213', 'eucjis2004',
    'euc_jisx0213', 'eucjisx0213',
    'euc_kr', 'euckr', 'korean', 'ksc5601', 'ks_c-5601', 'ks_c-5601-1987',
    'ksx1001', 'ks_x-1001',
    'gb2312', 'chinese', 'csiso58gb231280', 'euc-' 'cn', 'euccn',
    'eucgb2312-cn', 'gb2312-1980', 'gb2312-80', 'iso-' 'ir-58',
    'gbk', '936', 'cp936', 'ms936',
    'gb18030', 'gb18030-2000',
    'hz', 'hzgb', 'hz-gb', 'hz-gb-2312',
    'iso2022_jp', 'csiso2022jp', 'iso2022jp', 'iso-2022-jp',
    'iso2022_jp_1', 'iso2022jp-1', 'iso-2022-jp-1',
    'iso2022_jp_2', 'iso2022jp-2', 'iso-2022-jp-2',
    'iso2022_jp_2004', 'iso2022jp-2004', 'iso-2022-jp-2004',
    'iso2022_jp_3', 'iso2022jp-3', 'iso-2022-jp-3',
    'iso2022_jp_ext', 'iso2022jp-ext', 'iso-2022-jp-ext',
    'iso2022_kr', 'csiso2022kr', 'iso2022kr', 'iso-2022-kr',
    'latin_1', 'iso-8859-1', 'iso8859-1', '8859', 'cp819', 'latin', 'latin1',
    'L1',
    'iso8859_2', 'iso-8859-2', 'latin2', 'L2',
    'iso8859_3', 'iso-8859-3', 'latin3', 'L3',
    'iso8859_4', 'iso-8859-4', 'latin4', 'L4',
    'iso8859_5', 'iso-8859-5', 'cyrillic',
    'iso8859_6', 'iso-8859-6', 'arabic',
    'iso8859_7', 'iso-8859-7', 'greek', 'greek8',
    'iso8859_8', 'iso-8859-8', 'hebrew',
    'iso8859_9', 'iso-8859-9', 'latin5', 'L5',
    'iso8859_10', 'iso-8859-10', 'latin6', 'L6',
    'iso8859_13', 'iso-8859-13', 'latin7', 'L7',
    'iso8859_14', 'iso-8859-14', 'latin8', 'L8',
    'iso8859_15', 'iso-8859-15', 'latin9', 'L9',
    'iso8859_16', 'iso-8859-16', 'latin10', 'L10',
    'johab', 'cp1361', 'ms1361',
    'koi8_r', 'Russian',
    'koi8_u', 'Ukrainian',
    'mac_cyrillic', 'maccyrillic',
    'mac_greek', 'macgreek',
    'mac_iceland', 'maciceland',
    'mac_latin2', 'maclatin2', 'maccentraleurope',
    'mac_roman', 'macroman', 'macintosh',
    'mac_turkish', 'macturkish',
    'ptcp154', 'csptcp154', 'pt154', 'cp154', 'cyrillic-asian',
    'shift_jis', 'csshiftjis', 'shiftjis', 'sjis', 's_jis',
    'shift_jis_2004', 'shiftjis2004', 'sjis_2004', 'sjis2004',
    'shift_jisx0213', 'shiftjisx0213', 'sjisx0213', 's_jisx0213',
    'utf_32', 'U32', 'utf32',
    'utf_32_be', 'UTF-32BE',
    'utf_32_le', 'UTF-32LE',
    'utf_16', 'U16', 'utf16',
    'utf_16_be', 'UTF-16BE',
    'utf_16_le', 'UTF-16LE',
    'utf_7', 'U7', 'unicode-1-1-utf-7',
    'utf_8', 'U8', 'UTF', 'utf8',
    'utf_8_sig'])

# "Notice that spelling alternatives that only differ in case or use a hyphen
# instead of an underscore are also valid aliases; therefore, e.g. 'utf-8' is a
# valid alias for the 'utf_8' codec."
ALLOWED_ENCODINGS = set(item.lower().replace('_', '-')
                        for item in ALLOWED_ENCODINGS)

ParsedContentType = namedtuple(
    'ParsedContentType', 'content_type encoding params')


def get_content_type_info(environ,
                          default_type='application/octet-stream',
                          default_encoding='iso-8859-1'):
    """
    Read and parse the Content-Type header and return a
    :class:`ParsedContentType` object.
    """
    ct, params = parse_header(environ.get('CONTENT_TYPE', default_type))
    encoding = params.get('charset', default_encoding)
    if encoding is None or encoding.lower() not in ALLOWED_ENCODINGS:
        encoding = default_encoding
    return ParsedContentType(ct, encoding, params)


class TooBig(RequestParseError):
    """\
    Request body is too big
    """
    def __init__(self, *args, **kwargs):
        super(TooBig, self).__init__(*args, **kwargs)
        self.response = fresco.response.Response.request_entity_too_large()


class MissingContentLength(RequestParseError):
    """\
    No ``Content-Length`` header given
    """
    def __init__(self, *args, **kwargs):
        super(MissingContentLength, self).__init__(*args, **kwargs)
        self.response = fresco.response.Response.length_required()


def dequote(s):
    """
    Return ``s`` with surrounding quotes removed. Example usage:

        >>> from fresco.util.http import dequote
        >>> dequote('foo')
        'foo'
        >>> dequote('"foo"')
        'foo'
    """
    if len(s) > 1 and s[0] == '"' == s[-1]:
        return s[1:-1]
    return s


def parse_header(header):
    """
    Given a header, return a tuple of
    ``(value, [(parameter_name, parameter_value)])``.

    Example usage::

        >>> parse_header("text/html; charset=UTF-8")
        ('text/html', {'charset': 'UTF-8'})
        >>> parse_header("multipart/form-data; boundary=-------7d91772e200be")
        ('multipart/form-data', {'boundary': '-------7d91772e200be'})
    """
    if isinstance(header, Header):
        # Python3's email.parser.Parser returns a Header object (rather than
        # a string) for values containing 8-bit characters. These are then
        # replaced by U+FFFD when converting the header to a string
        header = ustr(header)

    elif isinstance(header, bytes):
        # In Python2 headers come through as raw byte strings. Use 'replace'
        # to emulate the PY3 behaviour:
        header = header.decode('ascii', 'replace')

    items = header.split(';')
    pairs = [(name, dequote(value))
             for name, value in (item.lstrip().split('=', 1)
                                 for item in items[1:])]
    return (items[0], dict(pairs))


def parse_querystring(data, charset=None, strict=False, keep_blank_values=True,
                      pairsplitter=re.compile('[;&]').split):
    """\
    Return ``(key, value)`` pairs from the given querystring::

        >>> list(parse_querystring('green%20eggs=ham;me=sam+i+am'))
        [(u'green eggs', u'ham'), (u'me', u'sam i am')]

    :param data: The query string to parse.
    :param charset: Character encoding used to decode values. If not specified,
                    ``fresco.DEFAULT_CHARSET`` will be used.

    :param keep_blank_values: if True, keys without associated values will be
                              returned as empty strings. if False, no key,
                              value pair will be returned.

    :param strict: if ``True``, a ``ValueError`` will be raised on parsing
                   errors.
    """

    if charset is None:
        charset = fresco.DEFAULT_CHARSET

    if PY3:
        unquote = lambda s: unquote_plus(s, charset)
    else:
        unquote = lambda s: unquote_plus(s).encode('iso-8859-1')\
                                           .decode(charset)

    for item in pairsplitter(data):
        if not item:
            continue
        try:
            key, value = item.split('=', 1)
        except ValueError:
            if strict:
                raise RequestParseError("bad query field: %r" % (item,))
            if not keep_blank_values:
                continue
            key, value = item, ''

        try:
            yield (unquote(key), unquote(value))
        except UnicodeDecodeError:
            raise RequestParseError("Invalid character data: can't decode"
                                    " as %r" % (charset,))


def parse_post(environ, fp, default_charset=None, max_size=16 * KB,
               max_multipart_size=2 * MB):
    """\
    Parse the contents of an HTTP POST request, which may be either
    application/x-www-form-urlencoded or multipart/form-data encoded.

    Returned items are either tuples of (name, value) for simple string values
    or (name, FileUpload) for uploaded files.

    :param max_multipart_size: Maximum size of total data for a multipart form
                               submission

    :param max_size: The maximum size of data allowed to be read into memory.
                     For a application/x-www-form-urlencoded submission, this
                     is the maximum size of the entire data. For a
                     multipart/form-data submission, this is the maximum size
                     of any individual field (except file uploads).
    """
    ct, charset, ct_params = get_content_type_info(
                                environ,
                                'application/x-www-form-urlencoded',
                                default_charset or fresco.DEFAULT_CHARSET)

    try:
        content_length = int(environ['CONTENT_LENGTH'])
    except (TypeError, ValueError, KeyError):
        raise MissingContentLength()

    try:
        if ct == 'application/x-www-form-urlencoded':
            if content_length > max_size:
                raise TooBig("Content Length exceeds permitted size")
            return parse_querystring(SizeLimitedInput(fp, content_length)
                                        .read()
                                        .decode('ASCII'), charset)
        else:
            if content_length > max_multipart_size:
                raise TooBig("Content Length exceeds permitted size")
            try:
                boundary = ct_params['boundary']
            except KeyError:
                raise RequestParseError(
                    "No boundary given in multipart/form-data content-type")
            return parse_multipart(SizeLimitedInput(fp, content_length),
                                boundary.encode('ASCII'), charset, max_size)
    except UnicodeDecodeError:
        raise RequestParseError("Payload contains non ascii data")


def get_body_bytes(environ, max_size=16 * KB):
    """
    Read a single message body from environ['wsgi.input'], returning a bytes
    object.
    """
    try:
        content_length = int(environ['CONTENT_LENGTH'])
    except (TypeError, ValueError, KeyError):
        raise MissingContentLength()

    if content_length > max_size:
        raise TooBig("Content Length exceeds permitted size")
    return SizeLimitedInput(environ['wsgi.input'], content_length).read()


def parse_body(environ, fp, default_charset=None, max_size=16 * KB):
    """
    Parse the message
        (payload as a byte string, content-type, encoding)
        """
    ct, charset, params = get_content_type_info(
        environ.get('CONTENT_TYPE', 'text/plain'))


class HTTPMessage(Message):
    """
    Represent HTTP request message headers
    """


def parse_multipart(fp, boundary, default_charset, max_size):
    """
    Parse data encoded as ``multipart/form-data``. Generate tuples of::

        (<field-name>, <data>)

    Where ``data`` will be a string in the case of a regular input field, or a
    ``FileUpload`` instance if a file was uploaded.

    :param fp: input stream from which to read data
    :param boundary: multipart boundary string, as specified by the
                     ``Content-Disposition`` header
    :param default_charset: character set to use for encoding, if not specified
                            by a content-type header. In practice web browsers
                            don't supply a content-type header so this needs to
                            contain a sensible value.
    :param max_size: Maximum size in bytes for any non file upload part
    """

    boundary_size = len(boundary)
    if not boundary.startswith(b'--'):
        raise RequestParseError("Malformed boundary string: "
                                "must start with '--' (rfc 2046)")

    if boundary_size > 72:
        raise RequestParseError("Malformed boundary string: "
                                "must be no more than 70 characters, not "
                                "counting the two leading hyphens (rfc 2046)")

    assert boundary_size + 2 < CHUNK_SIZE, \
          "CHUNK_SIZE cannot be smaller than the boundary string"

    if fp.read(2) != b'--':
        raise RequestParseError("Malformed POST data: expected two hypens")

    if fp.read(boundary_size) != boundary:
        raise RequestParseError("Malformed POST data: expected boundary")

    if fp.read(2) != b'\r\n':
        raise RequestParseError("Malformed POST data: expected CRLF")

    fp = PutbackInput(fp)

    while True:
        headers, data = _read_multipart_field(fp, boundary)
        try:
            _, params = parse_header(headers['Content-Disposition'])
        except KeyError:
            raise RequestParseError("Missing Content-Disposition header")

        try:
            name = params['name']
        except KeyError:
            raise RequestParseError("Missing name parameter in "
                                    "Content-Disposition header")

        is_file_upload = 'Content-Type' in headers and 'filename' in params

        if is_file_upload:
            storage = data.getstorage()
            storage.seek(0)
            yield name, FileUpload(params['filename'], headers, storage)

        else:
            charset = parse_header(headers.get('Content-Type', ''))[1]\
                        .get('charset', default_charset)
            if data.tell() > max_size:
                raise TooBig("Data block exceeds maximum permitted size")
            try:
                data.seek(0)
                yield name, data.read().decode(charset)
            except UnicodeDecodeError:
                raise RequestParseError("Invalid character data: can't decode "
                                        "as %r" % (charset,))

        chunk = fp.read(2)
        if chunk == b'--':
            if fp.peek(3) != b'\r\n':
                raise RequestParseError("Expected terminating CRLF "
                                        "at end of stream")
            break

        if chunk != b'\r\n':
            raise RequestParseError("Expected CRLF after boundary")


def _read_multipart_field(fp, boundary, parser=BytesParser(_class=HTTPMessage)):
    """
    Read a single part from a multipart/form-data message and return a tuple of
    ``(headers, data)``. Stream ``fp`` must be positioned at the start of the
    header block for the field.

    Return a tuple of ('<headers>', '<data>')

    ``headers`` is an instance of ``email.message.Message``.

    ``data`` is an instance of ``ExpandableOutput``.

    Note that this currently cannot handle nested multipart sections.
    """
    data = ExpandableOutput()
    headers = parser.parse(DelimitedInput(fp, b'\r\n\r\n'), headersonly=True)
    fp = DelimitedInput(fp, b'\r\n--' + boundary)

    # XXX: handle base64 encoding etc
    for chunk in iter(lambda: fp.read(CHUNK_SIZE), b''):
        data.write(chunk)
    data.flush()

    # Fallen off the end of the input without having read a complete field?
    if not fp.delimiter_found:
        raise RequestParseError("Incomplete data (expected boundary)")

    return headers, data


class FileUpload(object):
    """\
    Represent a file uploaded in an HTTP form submission
    """

    def __init__(self, filename, headers, fileob):

        self.filename = filename
        self.headers = headers
        self.file = fileob

        # UNC/Windows path
        if self.filename[:2] == '\\\\' or self.filename[1:3] == ':\\':
            self.filename = self.filename[self.filename.rfind('\\') + 1:]

    def save(self, fileob):
        """
        Save the upload to the file object or path ``fileob``

        :param fileob: a file-like object open for writing, or the path to the
                       file to be written
        """
        if isinstance(fileob, string_types):
            with open(fileob, 'wb') as f:
                return self.save(f)

        self.file.seek(0)
        copyfileobj(self.file, fileob)


def encode_multipart(data=None, files=None, charset='UTF-8', **kwargs):
    """
    Encode ``data`` using multipart/form-data encoding, returning a tuple
    of ``(<encoded data>, <environ items>)``.

    :param data: POST data to be encoded, either a dict or list of
                 ``(name, value)`` tuples.

    :param charset: Encoding used for any unicode values encountered in
                    ``data``

    :param files: list of ``(name, filename, content_type, data)`` tuples.
                    ``data`` may be either a byte string, iterator or
                    file-like object.

    :param kwargs: other data items as keyword arguments
    :returns: a tuple of ``(<encoded_data>, <environ_items>)``,
              where ``encoded_data`` is a BytesIO object
              and ``environ`` is a dict containing the Content-Type and
              Content-Length headers encoded for inclusion in a WSGI environ
              dict.
    """

    def header_block(name):
        return [('Content-Disposition', 'form-data; name="%s"' % (name,))]

    def file_header_block(name, filename, content_type):
        return [('Content-Disposition',
                 'form-data; name="%s"; filename="%s"' % (name, filename)),
                ('Content-Type', content_type)]

    def write_payload(stream, data):
        "Write ``data`` to ``stream``, encoding as required"
        if hasattr(data, 'read'):
            copyfileobj(data, stream)
        elif isinstance(data, bytes):
            stream.write(data)
        elif isinstance(data, string_types):
            stream.write(data.encode(charset))
        else:
            raise ValueError(data)

    if data is None:
        data = {}

    if files is None:
        files = []

    try:
        data = data.items()
    except AttributeError:
        pass

    data = chain(data, kwargs.items())

    boundary = b'-------' + hexlify(os.urandom(16))
    alldata = chain(
        ((header_block(k), payload) for k, payload in data),
        ((file_header_block(k, fn, ct), payload)
         for k, fn, ct, payload in files))

    CRLF = b'\r\n'
    post_data = BytesIO()
    post_data.write(b'--' + boundary)
    for headers, payload in alldata:
        post_data.write(CRLF)
        for name, value in headers:
            post_data.write('{0}: {1}\r\n'.format(name, value).encode('ascii'))
        post_data.write(CRLF)
        write_payload(post_data, payload)
        post_data.write(b'\r\n--' + boundary)
    post_data.write(b'--\r\n')
    length = post_data.tell()
    post_data.seek(0)
    wsgienv = {
        'CONTENT_LENGTH': str(length),
        'CONTENT_TYPE': unicode_to_environ(
                'multipart/form-data; boundary=' + boundary.decode('ascii'))}

    return (post_data, wsgienv)



from fresco import FrescoApp, GET
from fresco.exceptions import BadRequest, RedirectTemporary


class TestResponseExceptions(object):

    def test_exception_is_converted_to_response(self):

        def redirector():
            raise RedirectTemporary('/foo')

        app = FrescoApp()
        app.route('/', GET, redirector)

        with app.requestcontext('/'):
            assert app.view().status_code == 302

    def test_exception_can_have_its_response_modified(self):

        def view():
            raise BadRequest(content='custom error message',
                             content_type='application/foo')

        app = FrescoApp()
        app.route('/', GET, view)

        with app.requestcontext('/'):
            response = app.view()
            assert response.status_code == 400
            assert ''.join(response.content) == 'custom error message'
            assert response.get_header('Content-Type') == 'application/foo'



from fresco.util.cache import LRUCache


class TestLRUCache(object):

    lrucache = LRUCache

    def test_it_expires_least_recently_inserted(self):
        c = self.lrucache(3)
        c[0] = 0
        c[1] = 1
        c[2] = 2
        c[3] = 3
        assert set(c.keys()) == {1, 2, 3}

    def test_it_expires_least_recently_accessed(self):
        c = self.lrucache(3)
        c.update([(0, 0), (1, 1), (2, 2)])
        c.get(2)
        c.get(1)
        c.get(0)
        c[3] = 3
        assert set(c.keys()) == {0, 1, 3}

    def test_it_resizes_upwards(self):
        c = self.lrucache(2)
        c.update([(0, 0), (1, 1)])
        c.max_size = 4
        c.update([(2, 2), (3, 3), (4, 4)])
        assert set(c.keys()) == {1, 2, 3, 4}

    def test_it_resizes_downwards(self):
        c = self.lrucache(4)
        c.update([(0, 0), (1, 1), (2, 2), (3, 3)])
        c.max_size = 2
        assert set(c.keys()) == {2, 3}



from itertools import chain

from . import context, GET, Response
from .compat import ustr, string_types
from .routing import RouteArg
from .util.http import parse_header

__all__ = ['subrequest', 'subrequest_raw', 'subrequest_bytes']


class Markup(ustr):
    def __html__(self):
        return self


def subrequest(view, *args, **kwargs):
    """
    Return the response generated by a subrequest to ``view``

    This behaves differently depending on the value of ``view``:

    - **A string containing a ``/``**:
      a full request context is constructed to emulate a request
      to the given path.
      The subrequest will invoke all middleware layers and application hooks.

    - **A callable**: a new subrequest context will be created and
      ``view(*args, **kwargs)`` will be called. Middleware and application
      hooks will not be called, and any
      :class:`~fresco.routeargs.RouteArg`\s  defined will not be resolved.

    - **Any other value**: the view function is looked up using the same
      route resolution rules as :meth:`~fresco.core.FrescoApp.urlfor`.
      Middleware and hooks will be skipped, but
      :class:`~fresco.routeargs.RouteArg`\s will be resolved.

    If you pass in a view callable you can force RouteArgs to be resolved by
    specifying ``_resolve=True``.

    You can force a full request to be emulated by specifying ``_full=True``.

    Passing additional positional and keyword arguments has a different
    effect depending on the value of ``view``, ``_resolve`` and ``_full``:

    - **If view is a path, or if _full is True**,
      ``*args`` and ``**kwargs`` are passed to
      :meth:`~fresco.core.FrescoApp.requestcontext`,
      and may be used to customize the WSGI environment
      used for the subrequest.

    - **If view is a callable and _resolve is False**,
      ``*args`` and ``**kwargs`` are passed to the view callable.

    - **Otherwise**
      ``*args`` and ``**kwargs`` are passed to the route lookup mechanism
      and should provide the values defined in the route's path.
      ``kwargs`` values also override any :class:`~fresco.routeargs.RouteArg`\s
      defined by the route. These will only be generated from the request if
      a corresponding value is not present in ``kwargs``.

    :param view: The target path, view callable, or viewspec.

    :param _mode: One of ``raw``, ``str`` or ``bytes``.
                 If ``str``, the content of the subrequest will be
                 returned as a string.
                 If ``bytes``, the content will be returned as a
                 byte string.
                 If ``raw``, the raw response will be returned. In
                 this case the caller is responsible for iterating the
                 content and calling any functions in ``response.onclose``.

    :param _resolve: if ``True``, callable views will be resolved via the route
                     traversal mechanism, allowing routed arguments to be
                     set from the current request.

    :param _full: if ``True``, a full new request context will be set up
                  ensuring application hooks and middleware are called
                  for the subrequest. In this case any additional arguments
                  are passed to :meth:`~fresco.core.FrescoApp.requestcontext`
                  and used to populate the new request environ.

    :param args: Additional positional arguments

    :param kwargs: Additional keyword arguments
    """

    full = kwargs.pop('_full', False)
    resolve = kwargs.pop('_resolve', False)
    mode = kwargs.pop('_mode', 'str')

    if full or isinstance(view, string_types) and '/' in view:
        app = context.app
        if full:
            path = app.urlfor(view, *args, **kwargs)
        else:
            path = view
        with app.requestcontext(path, *args, **kwargs) as c:
            c.is_subrequest = True
            response = app.view()

    else:
        context.push(**dict(context.currentcontext(), is_subrequest=True))
        try:
            if not callable(view) or resolve:
                view, routed_args, routed_kwargs = \
                    resolve_viewspec(view, *args, **kwargs)

                # Use args/kwargs extracted during route traversal,
                args, kwargs = routed_args, routed_kwargs

            response = view(*args, **kwargs)
        finally:
            context.pop()

    if mode == 'str':
        return response_to_str(response)
    elif mode == 'bytes':
        return response_to_bytes(response)
    elif mode == 'raw':
        return response
    else:
        raise ValueError("Mode must be one of 'str', 'bytes', or 'raw'")


def subrequest_bytes(*args, **kwargs):
    """
    Perform a subrequest and return the response content as a byte string.
    See :func:`~fresco.subrequests.subrequest` for details
    """
    return subrequest(_mode='bytes', *args, **kwargs)


def subrequest_raw(*args, **kwargs):
    """
    Perform a subrequest and return the raw response object.
    See :func:`~fresco.subrequests.subrequest` for details
    """
    return subrequest(_mode='raw', *args, **kwargs)


def response_to_str(response):
    if not isinstance(response, Response):
        return response
    ct, params = parse_header(response.get_header('Content-Type'))
    encoding = params.get('charset')
    try:
        # Check for the common case that response.content is already a
        # string
        if isinstance(response.content, ustr):
            content = response.content

        # ...or a list of strings?
        elif (isinstance(response.content, list) and
                all(isinstance(i, ustr) for i in response.content)):
            content = ''.join(response.content)

        # Anything else: decode the content iterator
        else:
            content = b''.join(response.content_iterator)
            if not encoding:
                raise ValueError("No content encoding specified")
            content = content.decode(encoding)
    finally:
        if response.onclose:
            for f in response.onclose:
                f()
    if ct.startswith('text/html') or 'xml' in ct.split(';')[0]:
        return Markup(content)
    return content


def response_to_bytes(response):
    try:
        return b''.join(response.content_iterator)
    finally:
        if response.onclose:
            for f in response.onclose:
                f()


def resolve_viewspec(viewspec, *args, **kwargs):
    """
    :param viewspec: a view name, a reference in the form
    and used to
                        ``'package.module.viewfunction'``, or the view
                        callable itself.
    :param args: positional arguments for the route path traversal.
    :param kwargs: keyword arguments for the route path traversal.
    :return: A tuple of ``(<view callable>, args, kwargs)``.
             ```args`` and ``kwargs`` are the arguments to apply
             to ``view`` as extracted during the route traversal.
    """
    app = kwargs.pop('_routecollection', context.app)
    request = kwargs.pop('request', None) or context.request
    method = kwargs.pop('method', GET)
    args = list(args)

    if isinstance(viewspec, string_types) and ':' in viewspec:
        viewspec, remainder = viewspec.split(':', 1)
        delegated_route = app.routefor(viewspec)

        for k in delegated_route.routed_args_default:
            if k not in kwargs:
                v = delegated_route.routed_args_default[k]
                if callable(v):
                    v = v(request)
                kwargs[k] = v

        delegate_args, delegate_kwargs = _get_args_for_route(
            delegated_route, request, args, kwargs)

        rc = delegated_route.routecollectionfactory(*delegate_args,
                                                    **delegate_kwargs)
        return resolve_viewspec(remainder,
                                _routecollection=rc,
                                request=request,
                                method=method,
                                *args, **kwargs)

    route = app.routefor(viewspec)
    view = route.getview(method)
    args, kwargs = _get_args_for_route(route, request, args, kwargs)
    return view, args, kwargs


def _get_args_for_route(route, request, args, kwargs):
    """
    Return the args/kwargs required to pass to the view callable for ``route``.

    Mutates both args and kwargs
    """
    route_kwargs = {}
    route_args = list(chain(
        ((a(request) if isinstance(a, RouteArg) else a)
            for a in route.view_args),
        (args.pop(0) for a in route.pattern.segments
            if a.converter and not a.name)))

    required_kwargs = chain(route.view_kwargs.keys(),
                            (s.name
                             for s in route.pattern.segments
                             if s.converter and s.name))

    _marker = []
    for k in required_kwargs:
        v = kwargs.pop(k, _marker)
        if v is not _marker:
            route_kwargs[k] = v
            continue

        v = route.routed_args_default.get(k, _marker)
        if v is not _marker:
            if callable(v):
                v = v(request)
            route_kwargs[k] = v
            continue

        v = route.view_kwargs.get(k, _marker)
        if v is not _marker:
            v = v(request) if isinstance(v, RouteArg) else v
            route_kwargs[k] = v

    return route_args, route_kwargs



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from __future__ import absolute_import

from operator import attrgetter, itemgetter, methodcaller
from fresco.exceptions import MissingRouteArg
from itertools import cycle

__all__ = ('routearg', 'FormArg', 'PostArg', 'QueryArg', 'GetArg', 'CookieArg',
           'SessionArg', 'RequestObject',
           'FormData', 'PostData', 'QueryData', 'GetData')

_marker = []

magicmethods = [
    '__abs__', '__add__', '__and__', '__call__', '__cmp__', '__coerce__',
    '__complex__', '__contains__', '__del__', '__delattr__', '__delete__',
    '__delitem__', '__delslice__', '__div__', '__divmod__', '__enter__',
    '__eq__', '__exit__', '__float__', '__floordiv__',
    '__ge__', '__getattr__', '__getitem__', '__getslice__',
    '__gt__', '__hash__', '__hex__', '__iadd__', '__iand__', '__idiv__',
    '__ifloordiv__', '__ilshift__', '__imod__', '__imul__', '__index__',
    '__int__', '__invert__', '__iop__', '__ior__', '__ipow__', '__irshift__',
    '__isub__', '__iter__', '__itruediv__', '__ixor__', '__le__', '__len__',
    '__long__', '__lshift__', '__lt__', '__mod__', '__mul__', '__ne__',
    '__neg__', '__nonzero__', '__oct__', '__op__', '__or__', '__pos__',
    '__pow__', '__radd__', '__rand__', '__rcmp__', '__rdiv__', '__rdivmod__',
    '__repr__', '__reversed__', '__rfloordiv__', '__rlshift__', '__rmod__',
    '__rmul__', '__rop__', '__ror__', '__rpow__', '__rrshift__', '__rshift__',
    '__rsub__', '__rtruediv__', '__rxor__', '__setattr__', '__setitem__',
    '__setslice__', '__str__', '__sub__', '__truediv__', '__unicode__',
    '__weakref__', '__xor__']


def map_magics(dest):
    def map_magics(cls):
        dest_method = getattr(cls, dest)
        for item in magicmethods:
            setattr(cls, item, dest_method)
        return cls
    return map_magics


@map_magics('raise_exception')
class LazyException(object):
    """
    A lazy exception uses magic methods to intercept any access to the object
    and raise an exception.
    """

    def __init__(self, exception):
        self.__dict__['_exception'] = exception

    def raise_exception(self, *args, **kwargs):
        raise self.__dict__['_exception']


class RouteArg(object):
    """
    RouteArg objects can be used as keyword arguments in a route definition.
    RouteArgs can extract information from the request and make it available
    to the view.

    For example a RouteArg could be developed that reads information from the
    request cookie::

        Route('/', GET, myview, affiliate=CookieArg('affiliate_id'))

    A naive implementation of ``CookieArg`` could look like this::

        class CookieArg(RouteArg):

            def __init__(self, name):
                self.name = name

            def __call__(self, request):
                try:
                    return request.cookies[self.name]
                except KeyError,
                    return None

    When the route is constructed the RouteArg's ``configure`` method will
    be called with the Route object and the keyword name.

    At every request, the RouteArg instance will be called and expected to
    supply the value for its argument.
    """

    route = None
    name = None

    def __init__(self, default=_marker, exception=MissingRouteArg):
        self.default = default
        self.required = default is _marker
        self.exception = exception

    def configure(self, route, name):
        self.route = route
        self.name = name

    def __call__(self, request):
        return None


def routearg(func, *args, **kwargs):
    """
    Construct a ``RouteArg`` instance that calls ``func`` with the request
    object as an argument and passes the return value to the routed view::

        def extract_name(req):
            return req.get('name')

        Route('/', GET, myview, name=routearg(extract_name))

    Additional positional and keyword arguments may be passed, eg::

        def name_is(req, *friends):
            return req.get('name') in friends

        Route('/', GET, myview, is_a_friend=routearg(name_is, 'fred', 'jim'))

    """
    class _RouteArg(RouteArg):
        def __call__(self, request):
            return func(request, *args, **kwargs)
    return _RouteArg()


class RequestArg(RouteArg):
    """\
    Extract a view keyword argument from the request object.
    """

    query = attrgetter('query')
    form = attrgetter('form')
    cookies = attrgetter('cookies')
    session = attrgetter('session')

    #: Source for the request variable
    source = form

    #: Exceptions that signal the converter could not do its job due to invalid
    #: input
    converter_exceptions = (ValueError, TypeError)

    def __init__(self, converter=None, key=None,
                 default=_marker, exception=MissingRouteArg):
        super(RequestArg, self).__init__(default, exception)
        self.formkey = key
        self.is_list = isinstance(converter, list)
        if self.is_list:
            self.converter = lambda vs: [c(v)
                                         for c, v in zip(cycle(converter), vs)]
        else:
            self.converter = converter

    def configure(self, route, name):
        if self.formkey is None:
            self.formkey = name
        if self.is_list:
            self.getter = methodcaller('getlist', self.formkey)
        else:
            self.getter = itemgetter(self.formkey)

    def __call__(self, request):
        try:
            value = self.getter(self.source(request))
        except KeyError:
            if self.required:
                return LazyException(
                            self.exception('No value provided for %s' %
                                           (self.formkey,)))
            else:
                return self.default

        try:
            if self.converter is not None:
                value = self.converter(value)
        except self.converter_exceptions:
            value = LazyException(
                        self.exception('Invalid value for %s' %
                                       (self.formkey,)))

        return value


class RequestObject(RouteArg):
    """\
    Make the request object itself available as a view argument.

    Example::

        @app.route('/form', POST, request=RequestObject())
        def view(formdata, request):
            ...
    """
    def __call__(self, request):
        return request


class JSONPayload(RouteArg):
    """
    Make the JSON payload from the request available as a view argument.

    Example::

        @app.route('/form', POST, data=JSONPayload())
        def view(formdata, request):
            ...
    """
    def __call__(self, request):
        if request.content_type == 'application/json':
            return request.get_json()
        if self.default is not _marker:
            return self.default
        return LazyException(self.exception('JSON payload missing'))


class GetData(RouteArg):
    """\
    Make the ``request.form`` MultiDict available as a view argument.

    Example::

        @app.route('/form', POST, data=GetData())
        def view(formdata, data):
            ...
    """

    def __call__(self, request):
        return request.query


#: QueryData may be used as an alias for GetData
QueryData = GetData


class PostData(RouteArg):
    """\
    Make the ``request.form`` MultiDict available as a view argument.

    Example::

        @app.route('/form', POST, data=FormData())
        def view(formdata, data):
            ...
    """

    def __call__(self, request):
        return request.form

#: FormData may be used as an alias for PostData
FormData = PostData


class CookieArg(RequestArg):
    """\
    Extract a view keyword argument from ``request.cookies``.

    Example::

        @app.route('/', GET, message=CookieArg(key='msg', default=None))
        def view(message, message):
            return Response([message])
    """
    source = RequestArg.cookies

    def getter(self, cookies):
        if self.is_list:
            return cookies.getlist(self.formkey)
        else:
            return cookies[self.formkey]

    def configure(self, route, name):
        if self.formkey is None:
            self.formkey = name


class GetArg(RequestArg):
    """\
    Extract a view keyword argument from ``request.query``.

    Example::

        @app.route('/add', GET, a=QueryArg(int), b=QueryArg(int))
        def view(a, b):
            return Response(['a + b = %d' % (a + b)])
    """
    source = RequestArg.query

#: QueryArg may be used as an alias for GetArg
QueryArg = GetArg


class PostArg(RequestArg):
    """\
    Extract a view keyword argument from ``request.form``.

    Example::

        @app.route('/mul', POST, a=FormArg(int), b=FormArg(int))
        def view(a, b):
            return Response(['a * b = %d' % (a * b)])

    """
    source = RequestArg.form

#: FormArg may be used as an alias for PostArg
FormArg = PostArg


class SessionArg(RequestArg):
    """\
    Extract a view keyword argument from ``request.session``.

    Example::

        @app.route('/mul', POST, a=FormArg(int), b=FormArg(int))
        def view(a, b):
            return Response(['a * b = %d' % (a * b)])

    """
    source = RequestArg.session



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from __future__ import absolute_import

import sys

from collections import defaultdict
from contextlib import contextmanager
from functools import partial
import logging

from fresco.compat import reraise
from fresco.request import Request
from fresco.response import Response
from fresco.util.http import encode_multipart
from fresco.util.urls import normpath, make_query
from fresco.util.common import fq_path

from fresco.exceptions import ResponseException
from fresco.requestcontext import context
from fresco.routing import (GET, ExtensiblePattern, RouteCollection,
                            RouteNotFound, TraversedCollection)
from fresco.options import Options

__all__ = ('FrescoApp', 'urlfor')

logger = logging.getLogger(__name__)


class FrescoApp(RouteCollection):
    """\
    Fresco application class.
    """

    #: The default class to use for URL pattern matching
    pattern_class = ExtensiblePattern

    #: A stdlib logger object, or None
    logger = None

    #: Class to use to instantiate request objects
    request_class = Request

    def __init__(self, *args, **kwargs):

        from fresco.middleware import FrescoCoreMiddleware

        views = kwargs.pop('views', None)
        path = kwargs.pop('path', None)
        super(FrescoApp, self).__init__(*args, **kwargs)

        if views is not None:
            if path is None:
                path = '/'
            self.include(path, views)

        #: A list of (middleware, args, kwargs) tuples
        self._middleware = []

        #: Middleware layers applied after any middleware added through
        #: :meth:`add_middleware`.
        self.core_middleware = [partial(FrescoCoreMiddleware, frescoapp=self)]

        #: The WSGI application. This is generated when the first request is
        #: made.
        self._wsgi_app = None

        #: An options dictionary, for arbitrary application variables or
        #: configuration
        self.options = Options()

        #: Functions to be called before the routing has been traversed.
        #: Each function will be passed the request object.
        #: If a function returns a value (other than ``None``)
        #: this will be returned and the normal routing system bypassed.
        self.process_request_handlers = []

        #: Functions to be called after routing, but before any view is invoked
        #: Each function will be passed
        #: ``request, view, view_args, view_kwargs``.
        #: If a function returns a Response instance
        #: this will be returned as the response
        #: instead of calling the scheduled view.
        #: If a function returns any other (non-None) value this will be used
        #: to replace the scheduled view function.
        self.process_view_handlers = []

        #: Functions to be called after a response object has been generated
        #: Each function will be passed ``request, response``.
        #: If a function returns a value (other than ``None``),
        #: this value will be
        #: returned as the response instead of calling the scheduled view.
        self.process_response_handers = []

        #: Functions to be called if the response has an HTTP error status code
        #: (400 <= x <= 599)
        #: Each function will be passed ``request, response``.
        #: If a function returns a value (other than ``None``),
        #: this value will be
        #: returned as the response instead of calling the scheduled view.
        self.process_http_error_response_handlers = []

        #: Functions to be called if an exception is raised during a view
        #: Each function will be passed ``request, exc_info``.
        #: If a function returns a value (other than ``None``),
        #: this value will be
        #: returned as the response and the error will not be propagated.
        #: If all exception handlers return None then the error will be raised
        self.process_exception_handlers = []

        #: Functions to be called at the end of request processing,
        #: after all content has been output.
        #: Each function will be passed ``request`` and should not
        #: return any value.
        self.process_teardown_handlers = []

    def __call__(self, environ, start_response):
        """\
        Call the app as a WSGI application
        """
        if self._wsgi_app is None:
            self._wsgi_app = self._make_wsgi_app()
        return self._wsgi_app(environ, start_response)

    def __str__(self):
        """\
        String representation of the application and its configured routes
        """
        clsname = self.__class__.__name__
        return '<%s %s>' % (
            clsname,
            ('\n' + ' ' * (len(clsname) + 2)).join(str(r)
                                                   for r in self.__routes__))

    def get_response(
            self, routecollection=None, request=None, path=None, method=None):

        ctx = context.currentcontext()
        try:
            routecollection = routecollection or self
            request = request or ctx['request']
            method = method or request.method
            path = path or request.path_info
            error_response = None

            ctx['app'] = self
            environ = request.environ
            environ['fresco.app'] = self

            if path:
                path = normpath(path)
            else:
                path = '/'
        except ResponseException as e:
            return e.response

        response = None
        for f in self.process_request_handlers:
            try:
                r = f(request)
                if r is not None:
                    response = r
            except Exception:
                return self.handle_exception(request,
                                             allow_reraise=False)
        if response:
            return response

        try:
            for traversal in routecollection.get_routes(path, method, request):
                try:
                    route = traversal.route
                    environ['wsgiorg.routing_args'] = (traversal.args,
                                                       traversal.kwargs)
                    view = route.getview(method)
                    ctx['view_self'] = getattr(view, '__self__', None)
                    ctx['route_traversal'] = traversal
                    if self.logger:
                        self.logger.info("matched route: %s %r => %r",
                                         request.method, path, fq_path(view))

                    response = None
                    for f in self.process_view_handlers:
                        try:
                            r = f(request, view, traversal.args,
                                  traversal.kwargs)
                            if r is not None:
                                response = r
                        except Exception:
                            return self.handle_exception(request,
                                                         allow_reraise=False)
                    if response is not None:
                        if isinstance(response, Response):
                            return response
                        else:
                            view = response

                    view = route.getdecoratedview(view)
                    response = view(*traversal.args, **traversal.kwargs)

                except ResponseException as e:
                    if e.is_final:
                        return e.response
                    error_response = error_response or e.response

                except Exception:
                    return self.handle_exception(request)
                else:
                    return response

        except ResponseException as e:
            if e.is_final:
                return e.response
            error_response = error_response or e.response

        except Exception:
            return self.handle_exception(request)

        # A route was matched, but an error was returned
        if error_response:
            return error_response

        # Is this a head request?
        if method == 'HEAD':
            response = self.get_response(routecollection, request, path, GET)
            if '200' <= response.status < '300':
                return response.replace(content=[], content_length=0)
            return response

        # Is the URL matched by another HTTP method?
        methods = self.get_methods(routecollection, request, path)
        if methods:
            return Response.method_not_allowed(methods)

        # Is the URL just missing a trailing '/'?
        if path[-1] != '/':
            for _ in self.get_methods(routecollection, request, path + '/'):
                return Response.redirect_permanent(path + '/')

        return Response.not_found()

    def view(self, request=None):
        request = request or context.request
        response = self.get_response(request=request)

        for f in self.process_response_handers:
            try:
                r = f(request, response)
                if r is not None:
                    response = r
            except Exception:
                self.log_exception(request)

        if '400' <= response.status <= '599':
            if self.process_http_error_response_handlers:
                response = self.handle_http_error_response(request, response)

        return response

    def handle_http_error_response(self, request, response):
        """
        Call any process_http_error_response handlers and return the
        (potentially modified) response object.
        """
        for status, f in self.process_http_error_response_handlers:
            try:
                if status is not None and status != response.status_code:
                    continue
                r = f(request, response)
                if r is not None:
                    response = r
            except Exception:
                self.log_exception(request)
        return response

    def get_methods(self, routecollection, request, path):
        """\
        Return the HTTP methods valid in routes to the given path
        """
        methods = set()
        for traversal in routecollection.get_routes(path, None):
            route = traversal.route
            if route.predicate and not route.predicate(request):
                continue
            methods.update(route.methods)
        return methods

    def log_exception(self, request, exc_info=None):
        exc_info = exc_info or sys.exc_info()
        (self.logger or logger).error(
            "Exception in {0} {1}".format(request.method, request.url),
            exc_info=exc_info)

    def handle_exception(self, request, allow_reraise=True):

        exc_info = sys.exc_info()

        have_error_handlers = self.process_exception_handlers or any(
            st in (None, 500)
            for st, fn in self.process_http_error_response_handlers)

        # Backwards compatibility: if no exception or 500 http error
        # handlers have been installed we default to the old behavior
        # of raising the exception and letting the upstream
        # server handle it
        if allow_reraise and not have_error_handlers:
            reraise(exc_info)

        try:
            if not self.process_exception_handlers:
                self.log_exception(request, exc_info)
            response = None
            for exc_type, func in self.process_exception_handlers:
                try:
                    if not issubclass(exc_info[0], exc_type):
                        continue
                    r = func(request, exc_info)
                    if r is not None:
                        response = r
                except Exception:
                    self.log_exception(request)
            if response is not None:
                if isinstance(response, tuple) and len(response) == 3:
                    reraise(response)
                return response
            else:
                return Response.internal_server_error()
        finally:
            del exc_info

    def add_middleware(self, middleware, *args, **kwargs):
        """\
        Add a WSGI middleware layer

        Note that middleware is applied from the outside in. The first
        middleware added will occupy the innermost layer and be called last in
        each request cycle.
        """
        self._wsgi_app = None
        self._middleware.append((middleware, args, kwargs))

    def remove_middleware(self, middleware):
        """
        Remove the given WSGI middleware layer.

        :param middleware: A middleware object. This must be the an object
                           that was previouslly passed to
                           :meth:`add_middleware`.
        """
        self._wsgi_app = None
        self._middleware = [m for m in self._middleware
                            if m[0] is not middleware]

    def insert_middleware(self, position, middleware, *args, **kwargs):
        """
        Insert a middleware layer at the given position

        :param position: The index of the element before which to insert.
                         Note that middleware is applied from the outside in,
                         so middleware inserted at position zero will be called
                         last.
        :param middleware: A middleware callable.
        :param args: extra positional args to be passed to the middleware for
                     initialization
        :param kwargs: extra keyword args to be passed to the middleware for
                       initialization
        """
        self._wsgi_app = None
        self._middleware.insert(position, (middleware, args, kwargs))

    def _raw_wsgi_app(self, environ, start_response):
        request = self.request_class(environ)
        response = self.view(request)
        return response(environ, start_response)

    def _make_wsgi_app(self, app=None):
        app = app or self._raw_wsgi_app
        for middleware, args, kwargs in self._middleware:
            app = middleware(app, *args, **kwargs)
        for middleware in self.core_middleware:
            app = middleware(app)
        return app

    def urlfor(self, viewspec, *args, **kwargs):
        """\
        Return the url for the given view name or function spec

        :param viewspec: a view name, a reference in the form
                         ``'package.module.viewfunction'``, or the view
                         callable itself.
        :param _scheme: the URL scheme to use (eg 'https' or 'http').
        :param _netloc: the network location to use (eg 'localhost:8000').
        :param _script_name: the SCRIPT_NAME path component
        :param _query: any query parameters, as a dict or list of
                        ``(key, value)`` tuples.
        :param _fragment: a URL fragment to append.

        All other arguments or keyword args are fed to the ``pathfor`` method
        of the pattern.
        """
        scheme = kwargs.pop('_scheme', None)
        netloc = kwargs.pop('_netloc', None)
        query = kwargs.pop('_query', {})
        script_name = kwargs.pop('_script_name', None)
        fragment = kwargs.pop('_fragment', None)

        ctx = context.currentcontext()
        request = ctx['request']
        traversal = ctx.get('route_traversal')
        if traversal and traversal.collections_traversed:
            collections_traversed = traversal.collections_traversed
        else:
            collections_traversed = [
                TraversedCollection(self, '', None, (), {}, (), {})]

        exc = None
        for ct in collections_traversed[::-1]:
            try:
                path = ct.path + \
                        ct.collection.pathfor(viewspec, request=request,
                                           *args, **kwargs)
            except RouteNotFound as e:
                exc = e
                continue
            return request.make_url(scheme=scheme, netloc=netloc,
                                    SCRIPT_NAME=script_name,
                                    PATH_INFO=path,
                                    parameters='',
                                    query=query, fragment=fragment)
        raise exc or RouteNotFound(viewspec)

    def view_method_not_found(self, valid_methods):
        """
        Return a ``405 Method Not Allowed`` response.

        Called when a view matched the pattern but no HTTP methods matched.
        """
        return Response.method_not_allowed(valid_methods)

    def call_process_teardown_handlers(self, request):
        """
        Called once the request has been completed and the response content
        output.
        """
        for func in self.process_teardown_handlers:
            try:
                func(request)
            except Exception:
                self.log_exception(request)

    def process_request(self, func):
        """
        Register a ``process_request`` hook function
        """
        self.process_request_handlers.append(func)
        return func

    def process_view(self, func):
        """
        Register a ``process_view`` hook function
        """
        self.process_view_handlers.append(func)
        return func

    def process_response(self, func):
        """
        Register a ``process_response`` hook function
        """
        self.process_response_handers.append(func)
        return func

    def process_exception(self, func, exc_type=Exception):
        """
        Register a ``process_exception`` hook function
        """
        if isinstance(func, type) and issubclass(func, Exception):
            return partial(self.process_exception, exc_type=func)

        self.process_exception_handlers.append((exc_type, func))
        return func

    def process_http_error_response(self, func, status=None):
        """
        Register a ``process_http_error_response`` hook function
        """
        if isinstance(func, int):
            return partial(self.process_http_error_response, status=func)

        self.process_http_error_response_handlers.append((status, func))
        return func

    def process_teardown(self, func):
        """
        Register a ``process_teardown`` hook function
        """
        if self._wsgi_app:
            raise AssertionError("Cannot add hook: "
                                 "application is now receiving requests")
        self.process_teardown_handlers.append(func)

    @contextmanager
    def requestcontext(self, url='/', environ=None,
                       wsgi_input=b'', middleware=True, **kwargs):
        """
        Return the global :class:`fresco.requestcontext.RequestContext`
        instance, populated with a new request object modelling default
        WSGI environ values.

        Synopsis::

            >>> app = FrescoApp()
            >>> with app.requestcontext('http://www.example.com/view') as c:
            ...     print c.request.url
            ...
            http://www.example.com/view

        Note that ``url`` must be properly URL encoded.

        :param url: The URL for the request,
                    eg ``/index.html`` or ``/search?q=foo``.
        :param environ: values to pass into the WSGI environ dict
        :param wsgi_input: The input stream to use in the ``wsgi.input``
                           key of the environ dict
        :param middleware: If ``False`` the middleware stack will not be
                           invoked. Disabling the middleware can speed up
                           the execution considerably, but it will no longer
                           give an accurate simulation of a real HTTP request.
        :param kwargs: additional keyword arguments will be passed into the
                       WSGI request environment
        """

        from fresco.util.wsgi import WSGIStack
        from fresco.util.wsgi import make_environ

        if middleware:
            middleware = [
                lambda app, m=m, a=mw_args, k=mw_kwargs: m(app, *a, **k)
                for m, mw_args, mw_kwargs in self._middleware]

        else:
            middleware = []

        middleware.extend(self.core_middleware)
        kwargs.update(middleware=middleware)
        with WSGIStack(make_environ(url, environ, wsgi_input, **kwargs),
                       middleware):
            yield context

    def requestcontext_with_payload(
            self, url='/', data=None, environ=None, files=None,
            multipart=False, **kwargs):

        if files:
            multipart = True

        if multipart:
            wsgi_input, headers = encode_multipart(data, files)
            kwargs.update(headers)
        elif hasattr(data, 'read'):
            wsgi_input = data.read()
        elif isinstance(data, bytes):
            wsgi_input = data
        else:
            wsgi_input = make_query(data).encode('ascii')

        if 'CONTENT_LENGTH' not in kwargs:
            kwargs['CONTENT_LENGTH'] = str(len(wsgi_input))

        return self.requestcontext(
            url, environ, wsgi_input=wsgi_input, **kwargs)

    def requestcontext_post(self, *args, **kwargs):
        return self.requestcontext_with_payload(
            REQUEST_METHOD='POST', *args, **kwargs)

    def requestcontext_put(self, *args, **kwargs):
        kwargs['REQUEST_METHOD'] = 'PUT'
        return self.requestcontext_with_payload(*args, **kwargs)

    def requestcontext_patch(self, *args, **kwargs):
        kwargs['REQUEST_METHOD'] = 'PATCH'
        return self.requestcontext_with_payload(*args, **kwargs)

    def requestcontext_delete(self, *args, **kwargs):
        kwargs['REQUEST_METHOD'] = 'DELETE'
        return self.requestcontext(*args, **kwargs)


def collect_keys(items):
    """
    [(k1, v1), (k1, v2), (k2, v3)] -> [(k1, [v1, v2]), (k2, [v3])]
    """
    d = defaultdict(list)
    for key, value in items:
        d[key].append(value)
    return d.items()


def urlfor(*args, **kwargs):
    """\
    Convenience wrapper around :meth:`~fresco.core.FrescoApp.urlfor`.
    """
    return context.app.urlfor(*args, **kwargs)



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from __future__ import absolute_import

try:
    from urllib.parse import quote, unquote
except ImportError:
    from urllib import quote, unquote  # NOQA
from datetime import datetime, timedelta
import warnings

from .compat import ustr


class RequestCookie(ustr):

    @property
    def value(self):
        warnings.warn("The value attribute on request cookies is deprecated",
                      DeprecationWarning, stacklevel=2)
        return self


class Cookie(object):
    """
    Represents an HTTP cookie.

    See rfc2109, HTTP State Management Mechanism

    Example:

        >>> from fresco.cookie import Cookie
        >>> c = Cookie('session_id', 'abc123')
        >>> c.path = '/cgi-bin'
        >>> c.domain = '.example.org'
        >>> c.path
        '/cgi-bin'
        >>> str(c)
        'session_id=abc123;Domain=.example.org;Path=/cgi-bin'
    """
    attributes = [
        ("Comment", "comment"),
        ("Domain", "domain"),
        ("Expires", "expires"),
        ("Max-Age", "max_age"),
        ("Path", "path"),
        ("Secure", "secure"),
        ("HttpOnly", "httponly"),
    ]

    bool_attributes = set(['secure', 'httponly'])

    def __init__(
        self, name, value, max_age=None, expires=None, path='/',
        secure=False, domain=None, comment=None, httponly=False,
        version=1
    ):
        """
        Initialize a ``Cookie`` instance.
        """
        self.name = name
        self.value = value
        self.max_age = max_age
        self.path = path
        self.secure = secure
        self.domain = domain
        self.comment = comment
        self.version = version
        self.expires = expires
        self.httponly = httponly

    def __str__(self):
        """
        Returns a string representation of the cookie in the format, eg
        ``session_id=abc123;Path=/cgi-bin;Domain=.example.com;HttpOnly``
        """
        cookie = ['%s=%s' % (self.name, quote(str(self.value)))]
        for cookie_name, att_name in self.attributes:
            value = getattr(self, att_name, None)
            if att_name in self.bool_attributes:
                if value:
                    cookie.append(cookie_name)
            elif value is not None:
                if att_name == 'expires':
                    value = format_date(self._expires.utctimetuple())
                cookie.append('%s=%s' % (cookie_name, str(value)))
        return ';'.join(cookie)

    def set_expires(self, dt):
        """
        Set the cookie ``expires`` value to ``datetime`` object ``dt``
        """
        if dt is not None and not isinstance(dt, datetime):
            raise ValueError("%r is not a datetime" % (dt,))
        self._expires = dt

    def get_expires(self):
        """
        Return the cookie ``expires`` value as an instance of ``datetime``.
        """
        if self._expires is None and self.max_age is not None:
            if self.max_age == 0:
                # Make sure immediately expiring cookies get a date firmly in
                # the past.
                self._expires = datetime(1980, 1, 1)
            else:
                self._expires = datetime.now() + timedelta(seconds=self.max_age)

        return self._expires

    expires = property(get_expires, set_expires)


def expire_cookie(cookie_or_name, *args, **kwargs):
    """
    Synopsis:

        >>> from fresco import Response
        >>> from fresco.cookie import expire_cookie
        >>> def view():
        ...     return Response(set_cookie=expire_cookie('X', path='/'))
        ...
        >>> from fresco import FrescoApp
        >>> with FrescoApp().requestcontext() as c:
        ...     print(view().get_header('Set-Cookie'))
        ...
        X=;Expires=Tue, 01 Jan 1980 00:00:00 GMT;Max-Age=0;Path=/
    """
    if isinstance(cookie_or_name, Cookie):
        expire = cookie_or_name
    else:
        expire = Cookie(name=cookie_or_name, value='', *args, **kwargs)
    return Cookie(
        name=expire.name,
        value='',
        expires=datetime(1980, 1, 1),
        max_age=0,
        domain=kwargs.get('domain', expire.domain),
        path=kwargs.get('path', expire.path)
    )


def parse_cookie_header(cookie_string, unquote=unquote):
    """
    Return a list of cookie (name, value) pairs read from the request headers.

    :param cookie_string: The cookie, eg ``CUSTOMER=FRED``
    :param unquote: A function to decode values. By default values are assumed
                    to be url quoted. If ``None`` the raw value will be
                    returned
    """
    if not cookie_string:
        return []
    cookies = []

    for part in cookie_string.split(";"):

        try:
            k, v = part.strip().split("=", 1)
        except ValueError:
            continue

        if k[0] == '$':
            # An attribute (eg path or domain) pertaining to the most recently
            # read cookie.
            # These were defined in RFC2109 and RFC2965, but removed in RFC6265
            # (April 2011) and as far as I know not used by browsers.
            # In any case we are only interested in parsing the value,
            # so we can drop these.
            continue

        # Unquote quoted values ('"..."' => '...')
        if v and '"' == v[0] == v[-1] and len(v) > 1:
            v = v[1:-1]

        if unquote:
            k, v = unquote(k), unquote(v)
        cookies.append((k, RequestCookie(v)))

    return cookies


def format_date(utctimetuple):
    """
    Format a date for inclusion in a Set-Cookie header, eg
    'Sun, 06 Nov 1994 08:49:37 GMT'.

    According to RFC6265, this must be an
    "rfc1123-date, defined in RFC2616, Section 3.3.1"

    RFC2616 says in turn:

        HTTP applications have historically allowed three different formats
        for the representation of date/time stamps:

        Sun, 06 Nov 1994 08:49:37 GMT  ; RFC 822, updated by RFC 1123
        Sunday, 06-Nov-94 08:49:37 GMT ; RFC 850, obsoleted by RFC 1036
        Sun Nov  6 08:49:37 1994       ; ANSI C's asctime() format

        The first format is preferred as an Internet standard[...]

    """
    return '%s, %02d %s %04d %02d:%02d:%02d GMT' % (
        ('Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun')[utctimetuple[6]],
        utctimetuple[2],
        ('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct',
         'Nov', 'Dec')[utctimetuple[1] - 1],
        utctimetuple[0],
        utctimetuple[3],
        utctimetuple[4],
        utctimetuple[5],
    )
